#ifndef FEMMAPS_H
#define FEMMAPS_H

#include "fem.h"
#include <Eigen/Dense>

namespace fem
{
    
/// helper class for accessing gmsh jacobian vector as an Eigen Matrix3d
///     WARNING: this class does not copy the initial vector!
///
/// with Eigen, the jacobian is NOT transposed by default
///     it is returned "by column" by gmsh and Eigen uses column-major
///     storage by default
///     (!= C storage or numpy's default = row major)
///
///  gmsh =  [ element1, element2, element3, ... ]
///            element1 = [ gp1, gp2, gp3, ... ]
///                         gp1 = [ dx/du, dy/du, dz/du, dx/dv, ... ]
///
///                    [ dx/du dx/dv dx/dw ]
///  JacoMap(e, gp) =  [ dy/du dy/dv dy/dw ]
///                    [ dz/du dz/dv dz/dw ]
///
class JacoMap
{
    std::vector<double> &jacobians;
    size_t nelems; ///< only used to check bounds in debug - TODO: could be computed from jacobians.size()
    size_t nGP;

public:
    typedef Eigen::Map<Eigen::Matrix3d> M;

    JacoMap(std::vector<double> &_jacobians,
            size_t _nelems, size_t _nGP) : jacobians(_jacobians),
                                           nelems(_nelems), nGP(_nGP)
    {
    }
    M operator()(size_t e, size_t gp) const
    {
        assert(e < nelems && gp < nGP);
        return M(&jacobians[(9 * nGP) * e + 9 * gp]);
    }
};

/// helper class for accessing gmsh coord vector as an Eigen Vector3d
///     WARNING: this class does not copy the initial vector!
///
///  gmsh = [ elem1, elem2, elem3, ... ]
///           elem1 = [ gp1, gp2, gp3, ... ]
///                     gp1 = [ x, y, z ]
///
///  CoordMap(e, gp) = [ x ]
///                    [ y ]
///                    [ z ]

class CoordMap
{
    std::vector<double> &coord;
    size_t nelems; ///< only used to check bounds in debug - TODO: could be computed from coord.size()
    size_t nGP;

public:
    typedef Eigen::Map<Eigen::Vector3d> M;

    CoordMap(std::vector<double> &_coord,
             size_t _nelems, size_t _nGP) : coord(_coord),
                                            nelems(_nelems), nGP(_nGP)
    {
    }
    M operator()(size_t e, size_t gp) const
    {
        assert(e < nelems && gp < nGP);
        return M(&coord[(3 * nGP) * e + 3 * gp]);
    }
};

/// helper class for accessing gmsh dshapef vector as an Eigen MatrixXd
///
///    gmsh = [ gp1, gp2, gp2, ... ]
///             gp1 = [ shp1, shp2, shp3, ... ]
///                     shp1 = [df/du, df/dv, df/dw]
///
/// DShapeMap(gp) = [ df/du, df/dv, df/dw  ] <= shp1
///                 [ df/du, df/dv, df/dw  ] <= shp2
///                 [ df/du, df/dv, df/dw  ] <= shp3
///                 [ ...                  ]

class DShapeMap
{
    std::vector<double> &basisDF;
    size_t nGP; ///< only used to check bounds in debug - TODO: could be computed from basisDF.size()
    int nnod;

public:
    typedef Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>> M;

    DShapeMap(std::vector<double> &_basisDF,
              size_t _nGP, int _nnod) : basisDF(_basisDF), nGP(_nGP),
                                        nnod(_nnod)
    {
    }

    M operator()(size_t gp) const
    {
        assert(gp < nGP);
        return M(&basisDF[(nnod * 3) * gp], nnod, 3);
    }
};

} // namespace fem

#endif // FEMMAPS_H
